! Compute the 1-norm (col-sum-norm) and the max-norm (row-sum-norm) of a square matrix.
! In order to investigate the cache effects multiple versions of this norm-caclulation 
! has been written. The Fortran-sytle is used to allocate the matrix, so the matrix is 
! stored so-called column-wise.

!    The one norm of a (n x m) matrix is defined as:
!        ||A||_1        := max_{j=1..m} sum_{i=1}^n a_ij

!    The maximum norm of a matrix is defined as:
!        ||A||_infinity := max_{i=1..n} sum_{j=1}^m a_ij

module procedures
contains

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Initialise a n times n matrix
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine init (n,A)
implicit none

integer,intent(in) :: n
double precision,dimension(n,n),intent(out) :: A
integer :: i,j

do i=1, n-1
	do j=1, n-1
		A(i,j)=-1.
	end do
end do

do i=1,n
	A(n-1,i)=-i-2.
	A(i,n-1)=-i-2.
end do
end subroutine init

end module procedures

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! One norm of a matrix
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
function norm1(A,n)	
	implicit none
	integer,intent(in) :: n
	double precision,intent(in),dimension(n,n) :: A
	double precision :: norm1, colsum
	integer :: i,j
	colsum=0.
	norm1 = 0.
	do j=1, n
		colsum=0.
		do i=1, n
			colsum = colsum + abs (A(i,j))	
		end do
		if (norm1 < colsum) then
			norm1 = colsum
		end if
	end do
end function norm1

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Inifinty norm of a matrix (row-wise)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
function max_norm (A,n) 
	implicit none
	integer,intent(in) :: n
	double precision,intent(in),dimension(n,n) :: A
	double precision :: row_sum, max_norm
	integer :: i,j
	max_norm = 0.
	do i=1,n
		row_sum=0.
		do j=1,n
      !TODO: Calculate the row sum here. Hint: use the abs() function!
		end do
    !TODO: Update the max_norm, if the row sum is bigger
	end do
end function max_norm

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Infinity norm of a matrix (column-wise)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
function max_norm_col_wise (A,n)
	implicit none
	integer,intent(in) :: n
	double precision,intent(in),dimension(n,n) :: A
	double precision ::  max_norm_col_wise
	integer :: i,j
	double precision, dimension (n):: colwise_row_sum
  
  max_norm_col_wise = 0.

	!TODO: Init helper array with 0

	!TODO: Iterate the matrix in the correct order and 
  !calculate the row sums using the helper array

	!TODO: Search for the max_norm in the array colsums
 
end function max_norm_col_wise




program norm

use realtime
use procedures
	implicit none
	character(len = 128) :: arg
	integer :: n, length 
	double precision :: flops 
	double precision, allocatable, dimension(:,:) :: A
	integer,parameter :: num_tests = 5
	double precision :: n1, n2, n3, n4
	integer :: i,j,stat
	double precision :: t1, t2, t3, t4, time_start,tx
	double precision :: max_norm, norm1, max_norm_col_wise

  ! read command line parameter and allocate matrix
	call get_command_argument(1,arg,length,stat)
	if (stat==0) then
	 	read (arg,'(i10)') n
	else
		n = 5000
	end if
	write(*,*) 'n=', n
	
	allocate (A(n,n))
	
	call init(n,A)

	flops = n*n
		
	write (*,*) 'Compute 1-norm'
	time_start = GetRealTime()
	do i=1,num_tests
		n1=norm1(A,n)
	end do
	t1 = ( GetRealTime() - time_start ) / num_tests

	write (*,*) 'Compute maximum norm (row-wise)'
	time_start = GetRealTime()
	do i=1,num_tests
		n2=max_norm(A,n)
	end do
	t2 = ( GetRealTime() -time_start ) / num_tests
	
	write (*,*) 'Compute maximum norm (column-wise)'
	time_start = GetRealTime()
	do i=1,num_tests
		n3=max_norm_col_wise(A,n)
	end do
	t3 = ( GetRealTime() - time_start ) / num_tests
	
	! Compare results and print out the results and times
	write (*,*) 'The norms are:'
	
	write (*,'(A, E10.3, A, F8.3, A, F6.1, A)') '- 1-norm ', n1, ' in ', t1, ' sec (', flops/t1/1000000., ' Mflops)' 

	if (abs(n1-n2) <= epsilon(1.)) then 
    write (*,'(A, E10.3, A, F8.3, A, F6.1, A)') '- maximum norm (row)', n2, ' in ', t2, ' sec (', flops/t2/1000000., ' Mflops)'
	else
		write (*,*) '- 1-norm    FAILED: Wrong result!'
	end if

	if (abs(n1-n3) <= epsilon(1.)) then 
	 write (*,'(A, E10.3, A, F8.3, A, F6.1, A)') '- maximum norm (col) ', n3, ' in ', t3, ' sec (', flops/t3/1000000., ' Mflops)' 	
	else
		write (*,*) '- 1-norm    FAILED: Wrong result!'
	end if
	
end program norm 











