
module initialize
contains
 
subroutine init_data (m, n, y, A, x, ref)
integer, intent(in) :: m,n
double precision, dimension(n) :: x
double precision, dimension(m), intent (out) :: y,ref
double precision, dimension(n,m), intent(out) :: A
integer :: i,j
x=1
do i=1, m
	ref(i) = n*i
	y(i) = -1234.
	do j=1,n
		A(i,j) = j
	end do
end do
end subroutine init_data

end module initialize



module compute
contains

function check_results(m,n,y,ref)
	double precision, dimension(m),intent(in) :: y,ref
	integer, intent(in) :: m,n
	double precision :: relerr
	integer :: i  
	double precision :: TOL = epsilon (100.)
	double precision :: SMALL = epsilon (1.)
	check_results = 0	
	do i=1, m
		if (abs(ref(i)) > SMALL) then
			relerr = abs( (y(i)-ref(i)) / ref(i) )
		else
			relerr = abs((y(i)-ref(i)))
		end if
		if (relerr>TOL) then
			check_results = check_results + 1
		end if	
	end do

end function check_results

end module compute

module matr_vec
contains

subroutine mxv_ref (m,n,A,x,y)
implicit none
	integer, intent(in) :: m,n
	double precision, dimension(n,m), intent(in) :: A
	double precision, dimension(m), intent(out) :: y
	double precision, dimension(n), intent(in) :: x
	integer :: i,j
	
	do i=1, n
		y(i)=0.
		do j=1, m
      !TODO: Multiply x_j with a_ji and add the result to y_i
		end do
	end do
end subroutine mxv_ref
	
subroutine mxv_order (m,n,A,x,y)
implicit none
	integer, intent(in) :: m,n
	double precision, dimension(n,m), intent(in) :: A
	double precision, dimension(m), intent(out) :: y
	double precision, dimension(n), intent(in) :: x
	integer :: i,j

  !TODO: Initialize every y[i] with the first value!

  !TODO: Calculate the matrix-vector-product by accessing the matrix consecutively!

end subroutine mxv_order	

subroutine mxv_blas (m,n,A,x,y)
implicit none
	integer, intent(in) :: m,n
	double precision, dimension(n,m), intent(in) :: A
	double precision, dimension(m), intent(out) :: y
	double precision, dimension(n), intent(in) :: x
	double precision :: alpha, beta, incx, incy
	integer :: i,j

  alpha = 1
  beta = 0
  incx = 1
  incy = 1

  !TODO: Use BLAS such that dgemv() from the MKL is called and performs vxm!
end subroutine mxv_blas

end module matr_vec

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

program main
use omp_lib
use matr_vec
use initialize
use compute
implicit none
integer, parameter, dimension(7) :: dimensions = (/ 100, 250, 500, 1000, 2000, 4000, 5000 /)
integer :: ndimensions = 7, repeat_count = 30 
integer :: m,n,run
integer :: kr
double precision :: flops, t0, t1, t2, dt, memsize
double precision, allocatable, dimension(:,:) :: A
double precision, allocatable, dimension(:) :: x,y,ref
double precision :: tmp1,tmp2,tmp3

write(*,*)

write (*,*) '--------- values in Mflops --------- Ref --- Ordered --	'

do run=1, ndimensions
	n = dimensions(run)
	m = n
	flops = 2.*m*n
	memsize = 8. * (m + m*n + n)
	
	 allocate (A(n,m), x(n), y(m), ref(m))	
	 call init_data(m, n, y, A, x, ref)
	 
	
	
	 ! Execute and measure the reference solution
		t0 = omp_get_wtime()
		do kr=1, repeat_count
			call mxv_ref(m, n, A, x, y)
		end do	
		t1 = omp_get_wtime()
		dt = (t1-t0) / repeat_count;

		if(check_results(m, n, y, ref)>0) then
			tmp1 = -5.
		else
			tmp1 = (flops / dt) / 1000000.
		end if	


    ! Execute and measure the orderd solution
		t0 = omp_get_wtime()
		do kr=1, repeat_count
			call mxv_order(m, n, A, x, y)
		end do	
		t1 = omp_get_wtime()
		dt = (t1-t0) / repeat_count;

		if(check_results(m, n, y, ref)>0) then
			tmp2 = -5.
		else
			tmp2 = (flops / dt) / 1000000.
		end if	


    ! Execute and measure the BLAS solution
		t0 = omp_get_wtime()
		do kr=1, repeat_count
			call mxv_order(m, n, A, x, y)
		end do	
		t1 = omp_get_wtime()
		dt = (t1-t0) / repeat_count;

		if(check_results(m, n, y, ref)>0) then
			tmp3 = -5.
		else
			tmp3 = (flops / dt) / 1000000.
		end if	


	  write (*,*)
	  write (*,100) run, m, n,  memsize/1000000., tmp1, tmp2, tmp3
	  100 format (  '(',i1, '.)', i8,' x ',i8, ' ( ', f7.3, 'MB)',e10.3,e10.3,e10.3 )

	! cleanup .........
	 deallocate(A, x, y, ref)
	
end do
write (*,*) ''
end program main

